// Mail RSS items

#define _GNU_SOURCE

#include <err.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <assert_ops/libc.h>
#include <bestiola/assert.h>
#include <bestiola/compare.h>
#include <libxml/parser.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <system/sendmail.h>

enum {
	ADD_FEED_URL,
	GET_FEEDS
};

static char *configuration_dirname;
static char *feed_url;
static char *logname;
static int mode;

static void add_feed_url(void);
static const char *get_configuration_dirname(void);
static void get_feeds(void);
static char *getenv_copy(const char *name);
static char *mailsafe_string(const char *s);
static xmlDoc *open_configuration_file(void);
static void process_args(int argc, char **argv);
static void process_feed(int num, const char *url);

int main(int argc, char **argv) {
	process_args(argc, argv);

	switch (mode) {
	case ADD_FEED_URL:
		add_feed_url();
		break;
	case GET_FEEDS:
		get_feeds();
	}

	free(configuration_dirname);
	free(logname);

	return EXIT_SUCCESS;
}

static void abort_missing_arguments(void) {
	errx(EXIT_FAILURE, "missing arguments");
}

static void add_feed_url(void) {
	xmlDoc *document = open_configuration_file();
	if (document == NULL) {
		document = xmlNewDoc(xmlCast("1.0"));
		xmlNode *root = xmlNewNode(NULL, xmlCast("rss2mail-settings"));
		if (!root)
			assert_error_handler();
		xmlDocSetRootElement(document, root);
		xmlNode *node = xmlNewChild(root, NULL, xmlCast("feed"),
			xmlCast(feed_url));
		if (!node)
			assert_error_handler();
	}
	else
		warnx("Modifying configuration is not implemented");

	const char *dirname = get_configuration_dirname();
	char *path = assert_aprintf("%s/settings.xml", dirname);

	char *path_tmp = assert_aprintf("%s.tmp", path);

	mkdir(dirname, 0750);
	int ret = xmlSaveFormatFile(path_tmp, document, 1);
	if (ret == -1)
		assert_error_handler();

	// TODO: preserve file mode
	if (rename(path_tmp, path))
		assert_error_handler();
	free(path_tmp);
	free(path);

	xmlFreeDoc(document);
}

static xmlDoc *assert_xmlParseFile(const char *filename) {
	xmlDoc *document = xmlParseFile(filename);
	if (!document)
		assert_error_handler();
	return document;
}

static const char *get_configuration_dirname(void) {
	if (configuration_dirname)
		return configuration_dirname;

	char *home_dir = getenv_copy("HOME");
	if (!home_dir)
		assert_error_handler();
	configuration_dirname = assert_aprintf("%s/.rss2mail", home_dir);
	free(home_dir);
	return configuration_dirname;
}

static void get_feeds(void) {
	xmlDoc *document = open_configuration_file();
	if (document == NULL)
		errx(EXIT_FAILURE, "Could not parse configuration file.");

	// TODO: validate document, use schema

	int num = 0;
	xmlNode *node = xmlDocGetRootElement(document);
	node = node->children;
	for (; node; node = node->next) {
		if (xmlStrEqual(node->name, xmlCast("feed")))
			process_feed(num++, xmlUncast(node->children->content));
	}
	xmlFreeDoc(document);
}

static const char *get_logname(void) {
	if (logname)
		return logname;

	logname = getenv_copy("LOGNAME");
	if (!logname)
		assert_error_handler();
	return logname;
}

static char *getenv_copy(const char *name) {
	static pthread_mutex_t env_mutex = PTHREAD_MUTEX_INITIALIZER;
	pthread_mutex_lock(&env_mutex);
	char *copy = NULL;
	char *value = getenv(name);
	if (value)
		copy = strdup(value);
	pthread_mutex_unlock(&env_mutex);
	return copy;
}

static void mail_item(const char *channel_title, const xmlNode *item) {
	char *title = NULL;
	const char *link = NULL;
	const char *description = NULL;
	const char *date = NULL;
	xmlNode *node = item->children;
	for (; node; node = node->next) {
		if (xmlStrEqual(node->name, xmlCast("title"))) {
			title = mailsafe_string(xmlUncast(
				node->children->content));
			node = node->next;
			break;
		}
	}
	for (; node; node = node->next) {
		if (xmlStrEqual(node->name, xmlCast("link"))) {
			link = xmlUncast(node->children->content);
			node = node->next;
			break;
		}
	}
	for (; node; node = node->next) {
		if (xmlStrEqual(node->name, xmlCast("description"))) {
			description = xmlUncast(node->children->content);
			node = node->next;
			break;
		}
	}
	for (; node; node = node->next) {
		if (xmlStrEqual(node->name, xmlCast("date"))) {
			date = xmlUncast(node->children->content);
			break;
		}
	}
	if (!date) {
		warnx("Unexpected item format");
		free(title);
		return;
	}
	int year;
	int month;
	int day;
	int ret = sscanf(date, "%d-%d-%d", &year, &month, &day);
	if (ret != 3)
		warnx("Unexpected date format");
	struct tm stamp;
	stamp.tm_sec = 0;
	stamp.tm_min = 0;
	stamp.tm_hour = 0;
	stamp.tm_mday = day;
	stamp.tm_mon = month - 1;
	stamp.tm_year = year - 1900;
	stamp.tm_isdst = 0;
	if (mktime(&stamp) == -1)
		warnx("Timestamp error");
	static const char *DAYNAME[] = { "Sun", "Mon", "Tue", "Wed", "Thu",
		"Fri", "Sat" };
	static const char *MONTHNAME[] = { "Jan", "Feb", "Mar", "Apr", "May",
		"Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
	const char *user = get_logname();
	char *msg = assert_aprintf("\
From %s@localhost Thu Jan  1 00:00:00 UTC 1970\n\
From: %s <noreply@invalid>\n\
To: %s@localhost\n\
Subject: %s\n\
MIME-Version: 1.0\n\
Content-Type: text/html; charset=UTF-8\n\
Content-Transfer-Encoding: binary\n\
Date: %s, %02d %s %d 00:00:00 +0000\n\
\n\
%s\n\
<a href=\"%s\">%s</a>\n", user, channel_title, user, title,
	DAYNAME[stamp.tm_wday], day, MONTHNAME[stamp.tm_mon], year, description,
	link, link);
	free(title);
	if (system_sendmail(msg))
		assert_error_handler();
	free(msg);
}

static char *mailsafe_string(const char *s) {
	char *nl = strchrnul(s, '\n');
	size_t len = nl - s;
	char *safe = assert_malloc(len + 1);
	memcpy(safe, s, len);
	safe[len] = '\0';
	return safe;
}

static xmlDoc *open_configuration_file(void) {
	char *path = assert_aprintf("%s/settings.xml",
		get_configuration_dirname());
	xmlDoc *document = NULL;
	if (!access(path, F_OK))
		document = assert_xmlParseFile(path);
	free(path);
	return document;
}

static void process_args(int argc, char **argv) {
	switch (argc) {
	case 1:
		abort_missing_arguments();
	case 2:
		if (!strcmp(argv[1], "add"))
			abort_missing_arguments();
	}

	if (!strcmp(argv[1], "add")) {
		mode = ADD_FEED_URL;
		feed_url = argv[2];
	}
	else if (!strcmp(argv[1], "run"))
		mode = GET_FEEDS;
	else
		errx(EXIT_FAILURE, "unrecognized command %s", argv[1]);
}

static void process_feed(int num, const char *url) {
	const char *dirname = get_configuration_dirname();
	char *path = assert_aprintf("%s/%d.xml", dirname, num);

	// TODO: optimization
	// try libsoup to avoid redownload, issue a HEAD, check size/date
	// implement download to a file
	pid_t pid = fork();
	switch (pid) {
	case -1:
		assert_error_handler();
	case 0:
		if (!freopen("/dev/null", "r", stdin)
			|| !freopen("/dev/null", "w", stdout)
			|| !freopen("/dev/null", "w", stderr))
			exit(EXIT_FAILURE);
		static const char WGET[] = "wget";
		execlp(WGET, WGET, "-O", path, url, NULL);
		exit(EXIT_FAILURE);
	}

	siginfo_t info;
	if (waitid(P_PID, pid, &info, WEXITED))
		assert_error_handler();
	if (info.si_code != CLD_EXITED)
		assert_error_handler();
	if (info.si_status != EXIT_SUCCESS) {
		static const char *WGET_ERROR[] = {
			"Generic error code",
			"Parse error",
			"File I/O error",
			"Network failure",
			"SSL verification failure",
			"Username/password authentication failure",
			"Protocol errors",
			"Server issued an error response",
			"Unknown error"
		};
		unsigned error_idx = info.si_status - 1;
		if (error_idx > 8)
			error_idx = 8;
		errx(EXIT_FAILURE, "wget: %s.", WGET_ERROR[error_idx]);
	}

	char *path_last = assert_aprintf("%s/%d.last.xml", dirname, num);

	// TODO: There are several rss formats
	int old_count = 0;
	char **old_refs = NULL;
	if (!access(path_last, F_OK)) {
		xmlDoc *document = assert_xmlParseFile(path_last);
		xmlNode *root = xmlDocGetRootElement(document);
		xmlNode *node;
		for (node = root->children; node; node = node->next) {
			if (xmlStrEqual(node->name, xmlCast("item")))
				old_count++;
		}
		old_refs = assert_malloc(old_count * sizeof(char *));
		int i = 0;
		for (node = root->children; node; node = node->next) {
			if (!xmlStrEqual(node->name, xmlCast("item")))
				continue;
			old_refs[i++] = xmlUncast(xmlGetProp(node,
				xmlCast("about")));
		}
		xmlFreeDoc(document);
	}
	qsort(old_refs, old_count, sizeof(char *), compar_strcmp);

	xmlDoc *document = assert_xmlParseFile(path);
	xmlNode *root = xmlDocGetRootElement(document);
	char *channel_title = NULL;
	xmlNode *node = root->children;
	for (; node; node = node->next) {
		if (!xmlStrEqual(node->name, xmlCast("channel")))
			continue;
		xmlNode *node2 = node->children;
		for (; node2; node2 = node2->next) {
			if (!xmlStrEqual(node2->name, xmlCast("title")))
				continue;
			channel_title = mailsafe_string(xmlUncast(
				node2->children->content));
			break;
		}
		node = node->next;
		break;
	}
	for (; node; node = node->next) {
		if (!xmlStrEqual(node->name, xmlCast("item")))
			continue;
		char *ref = xmlUncast(xmlGetProp(node, xmlCast("about")));
		void *found = bsearch(&ref, old_refs, old_count, sizeof(char *),
			compar_strcmp);
		free(ref);
		if (!found)
			mail_item(channel_title, node);
	}
	free(channel_title);
	xmlFreeDoc(document);

	int i;
	for (i = 0; i < old_count; i++)
		free(old_refs[i]);
	free(old_refs);

	if (rename(path, path_last))
		assert_error_handler();
	free(path);
	free(path_last);
}
